/*
  libjlc.h - "C" API prototypes for the James Ladd Code Library.
 
  Copyright (c) 2006 James Ladd.
  http://www.hyperserve.org

  Please read the LICENSE file.
  
  Conventions:

  . All jameladdcode function names start with 'jlc_' and take the form
    jlc_<noun>_<verb>(). However, in some cases the <noun> is not specified.


  . All jlc functions return zero (0) if there are no errors during
    the execution of the function and the results are usable. Otherwise the
    value returned identifies the error condition(s).

  . An attempt has been made to keep parameters consistent (order) and
    uniform (size). For example, the parameters to API's will almost 
    always be the same size as a machine register (32bits) regardless of
    their use.

  . The API comments are organised into the following paragraphs:
    . function name
    . purpose
    . parameters
    . return
    . comments
    
  . The API is designed to provide the basic and minimum implementation
    necessary to be of use. Additional functionality may be added later.
    
  . Portions of this code use or are based upon the following works:

    Paul Brennick and Erol Ojalvo
    The GeneSys Project
    http://x86assembly.codegurus.org/
    
    AND
    
    Maged M. Michael and Michael L. Scott.
    Non-Blocking Algorithms and Preemption-Safe Locking on Multiprogrammed
    Shared Memory Multiprocessors.
    Manuscript submitted for journal publication, March 1997.
    http://www.cs.rochester.edu/u/michael/PODC96.html

*/

#ifdef __cplusplus
extern "C" {
#endif 

/*
 * Types required by this library.
 */
typedef unsigned int jlc_uint;
typedef unsigned int jlc_lib;
typedef unsigned int jlc_tpool;


extern char* cpuidfunc(void) asm ("_cpuidfunc");

/*
  jlc_lib_attach

  Attach and initialize the James Ladd Code library ready for use by the 
  process or thread making the call.

  jlclib - Pointer to the memory location where the address of the 
           attached library environment/context should be stored.

  Return:  Zero if there are no errors attaching to the library, otherwise
           a non-zero value indicating the error that occurred.

  A library attached in error should not be detached and should not be
  used.

  A thread or process that wants to make use of the API's in this library
  should call jlc_lib_attach once to initialize the library and
  establish the environment/context for the other API's.

  Threads within a process can share the jlclib pointer but different
  processes should not.

  Threads within a process can make new a environment/context for their
  use by calling jlc_lib_attach again.

  Each call to jlc_lib_attach should be matched with a call to 
  jlc_lib_detach when the environment/context is no longer required.
*/ 
extern jlc_uint jlc_lib_attach(jlc_lib* jlclib) asm ("_jlc_lib_attach");

/*
  jlc_lib_detach
 
  Detach and cleanup the library environment/context.

  jlclib - Pointer to the attached jlc library.

  Return:  Zero if there are no errors closing the thread pool,
           otherwise a non-zero value indicating the error that occurred.

  The library is detached, closed and all resources are released. A library
  pointer is invalid after this API returns.

  There should be a call to jlc_lib_detach for each call that was made to
  jlc_lib_attach.
*/
extern jlc_uint jlc_lib_detach(jlc_lib* jlclib) asm ("_jlc_lib_detach");

/*
  jlc_tpool_callback

  Defines the signature and arguments to the thread pool task callback function.
  This function is called by a worker thread in the pool when a task is to be
  serviced/run.
  
  jlclib - Handle to the attached jlc library.

  tpool  - Handle to the opened thread pool.

  val1   - Value that was passed when the task was submitted.

  val2   - Value that was passed when the task was submitted.

  thread - Handle to the worker thread that is servicing the task.
           
  Return:  Zero if there are no errors running the task, otherwise a non-zero
           value indicating the error that occurred.

  It is recommended that this function be as quick as possible with long tasks
  broken into smaller ones that are submitted after the current task
  completes.
  
  The signature of the callback is:
  
  jlc_uint jlc_tpool_callback(jlc_lib jlclib, jlc_tpool tpool, jlc_uint val1, 
                              jlc_uint val2, void* thread);
*/

/*
  jlc_tpool_open

  Open a thread pool of <tcount> threads to process tasks from
  a task queue of maximum <qcount> tasks. An open thread pool will not
  process tasks until it is running. See the jlc_tpool_run API.

  jlclib - Handle to the attached jlc library.

  tpool  - Pointer to the memory location where the address of the opened
           thread pool should be stored.

  tcount - The maximum count of threads in the thread pool.

  qcount - The maximum count of tasks that can be queued waiting to be
           serviced by a thread in the thread pool.

  Return:  Zero if there are no errors opening the thread pool,
           otherwise a non-zero value indicating the error that occurred.

  A thread pool opened without error should be closed with the
  function jlc_thread_close when no longer in use. A thread pool opened
  with an error should not be used and should not be closed.

  Tasks can be submitted to the thread pool even if the task queue is full.
  In this situation the process submitting the task must wait for the task
  queue to be serviced by a thread for a free task queue slot to become
  available.

  Tasks can be submitted to the thread pool even if the pool is not yet
  running. In this situation the tasks are placed on the task queue but will
  not be handled until the thread pool is running. See the jlc_tpool_run
  API for more information.

  It is recommended but not mandatory that more than one thread pool be
  created to handle tasks in a multi-threaded application, with a new thread
  pool created to handle tasks in different contexts. For example, one
  thread pool to service IO events and another to handle GUI events.

  It is recommended that tasks serviced by the thread pool run for as short
  a time as possible with larger tasks broken down into smaller tasks
  that can be submitted when a preceeding task is complete.

  The implementation of the thread pool may create all the threads and
  allocate all the memory necessary for the task queue at the time the
  thread pool is opened.
*/
extern jlc_uint jlc_tpool_open(jlc_lib jlclib, jlc_tpool* tpool, jlc_uint tcount,\
	                jlc_uint qcount) asm ("_jlc_tpool_open");

/*
  jlc_tpool_run

  Run a thread pool by signaling the worker threads that it's ok to start
  processing task submissions. 

  tpool  - Handle to the opened thread pool.

  Return:  Zero if there are no errors signaling the run of the thread pool,
           otherwise a non-zero value indicating the error that occurred.
	   
  A thread pool that has been opened but is not yet running can accept task
  submissions but they will not be handled by worker threads until the thread
  pool is run.

  A call to jlc_tpool_run that doesn't return zero indicates that the thread
  pool should be closed and not used.
*/
extern jlc_uint jlc_tpool_run(jlc_tpool tpool) asm ("_jlc_tpool_run");

/*
  jlc_tpool_close

  Close and cleanup a thread pool.

  tpool  - Pointer to the memory location where the address of the opened
           thread pool is stored.

  drain  - Non-zero indicates draining required.
  
  millis - Time in milliseconds to wait for the pool to be drained.

  Return:  Zero if there are no errors closing the thread pool,
           otherwise a non-zero value indicating the error that occurred.
  
  The thread pool is closed and all resources are released. A thread pool
  pointer is invalid after this API returns.
  
  A <drain>ed pool will wait for each task in the pool to be serviced before
  closing and this may take some time to complete depending on the type and
  count of the tasks queued.
  
  A <millis> that is non-zero indicated the time in milli-seconds that the
  pool will wait for all tasks to be drained before the API returns. If 
  <drain> is zero then <millis> as no effect and the pool will not be
  drained before it is closed. ie: the pool is forceably closed.
  
  A closed thread pool won't accept new task submissions.
  
  The initial implementation of the thread pool may not support the <millis>
  functionality and you should read the release notes to see when this
  is implemented.
*/
extern jlc_uint jlc_tpool_close(jlc_tpool* tpool, jlc_uint drain,\
	       		jlc_uint millis) asm ("_jlc_tpool_close");

/*
  jlc_tpool_submit

  Submit a task to the pool for servicing by a worker thread.

  tpool    - Handle to the opened thread pool.

  val1     - A value that is passed onto the task callback.
  
  val2     - A value that is passed onto the task callback.
  
  callback - Pointer to the task function.

  Return:  Zero if there are no errors submitting the task,
           otherwise a non-zero value indicating the error that occurred.

  A task is submitted to the thread pool by placing it on the queue of tasks
  to be serviced by worker threads. When a worker thread is free it will 
  service the task by calling the <callback> and passing it <val1> and <val2>
  as well as other values. See the jlc_tpool_task API for details on the
  signature of the callback function and what arguments are passed to it.
  
  When this API returns with a zero value the task is queued.
 
  This API will not return until the task is queued. Special attention has 
  been payed to the design of the queue to ensure it is thread safe and 
  non-blocking. This should ensure that the time spent queueing tasks is as
  small as possible.

  An open pool will accept task submissions.
  An open but not running pool will accept task submissions.
  A closed thread pool won't accept task submissions.
*/
extern jlc_uint jlc_tpool_submit(jlc_tpool tpool, jlc_uint val1, jlc_uint val2,\
	       		void* callback) asm ("_jlc_tpool_submit");

#ifdef __cplusplus
}
#endif

